/*
 * hauppauge IR driver
 *      (c) 2000 Gerd Knorr <kraxel@goldbach.in-berlin.de>
 *
 * parts are cut&pasted from lirc_haup.c
 *
 * this driver is useful only with the lirc package, see
 *	http://fsinfo.cs.uni-sb.de/~columbus/lirc/
 *
 * it will be moved back to the lirc package in near future.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/i2c.h>
#include <linux/videodev.h>
#include <asm/semaphore.h>

#ifdef __SMP__
#include <asm/pgtable.h>
#include <linux/smp_lock.h>
#endif
/* kernel_thread */
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) 
# include "kcompat24.h"
#endif
#include "lirc.h"

/* Addresses to scan */
static unsigned short normal_i2c[] = {I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {0x18,0x1a,I2C_CLIENT_END};
static unsigned short probe[2]        = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short probe_range[2]  = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore[2]       = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore_range[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short force[2]        = { I2C_CLIENT_END, I2C_CLIENT_END };
static struct i2c_client_address_data addr_data = {
	normal_i2c, normal_i2c_range, 
	probe, probe_range, 
	ignore, ignore_range, 
	force
};

#define BUFLEN 256
#define IR_MAX 4

struct IR {
	struct task_struct  *thread;
	struct semaphore    *notify;
	int                 rmmod;
	int                 open;

	struct semaphore    lock;
	wait_queue_head_t   wait_poll;
	unsigned short buffer[BUFLEN];
	int head,tail;
};

static struct i2c_client *irs[IR_MAX];

/* ----------------------------------------------------------------------- */
/* insmod parameters                                                       */

static int debug   = 0;    /* debug output */

MODULE_PARM(debug,"i");

#define dprintk if (debug) printk

/* ----------------------------------------------------------------------- */

#define DEVICE_NAME "hauppauge ir rc"
#define CODE_LENGTH 13
#define LIRC_MAJOR  61

/*
 * If this key changes, a new key was pressed.
 */
#define REPEAT_TOGGLE_0      192
#define REPEAT_TOGGLE_1      224

/*
 * KEY SEQUENCES.
 * Hauppauge uses these definitions internally in transmitting keypresses.
 * BYTE 2 contains the last key pressed.
 */
#define REMOTE_0           0x00
#define REMOTE_1           0x04
#define REMOTE_2           0x08
#define REMOTE_3           0x0c
#define REMOTE_4           0x10
#define REMOTE_5           0x14
#define REMOTE_6           0x18
#define REMOTE_7           0x1c
#define REMOTE_8           0x20
#define REMOTE_9           0x24
#define REMOTE_RADIO       0x30
#define REMOTE_MUTE        0x34
#define REMOTE_TV          0x3c
#define REMOTE_VOL_PLUS    0x40
#define REMOTE_VOL_MINUS   0x44
#define REMOTE_RESERVED    0x78
#define REMOTE_CHAN_PLUS   0x80
#define REMOTE_CHAN_MINUS  0x84
#define REMOTE_SOURCE      0x88
#define REMOTE_MINIMIZE    0x98
#define REMOTE_FULL_SCREEN 0xb8

/* ----------------------------------------------------------------------- */

static char *
keyname(unsigned char v)
{
	switch (v) {
	case REMOTE_0:              return "0";
	case REMOTE_1:              return "1";
	case REMOTE_2:              return "2";
	case REMOTE_3:              return "3";
	case REMOTE_4:              return "4";
	case REMOTE_5:              return "5";
	case REMOTE_6:              return "6";
	case REMOTE_7:              return "7";
	case REMOTE_8:              return "8";
	case REMOTE_9:              return "9";
	case REMOTE_RADIO:          return "Radio";
	case REMOTE_MUTE:           return "Mute";
	case REMOTE_TV:             return "TV";
	case REMOTE_VOL_PLUS:       return "Vol +";
	case REMOTE_VOL_MINUS:      return "Vol -";
	case REMOTE_RESERVED:       return "Reserved";
	case REMOTE_CHAN_PLUS:      return "Chan +";
	case REMOTE_CHAN_MINUS:     return "Chan -";
	case REMOTE_SOURCE:         return "Source";
	case REMOTE_MINIMIZE:       return "Minimize";
	case REMOTE_FULL_SCREEN:    return "Full Screen";
	default:    dprintk("Error, invalid key %d",v);
		return "";
	}
}

static int
ir_poll_thread(void *data)
{
	struct i2c_client *client = data;
	struct IR *ir = client->data;
        unsigned char b[3], last_b0 = 0;
	__u16 repeat_bit, code;
    
#ifdef __SMP__
	lock_kernel();
#endif

	exit_mm(current);
	current->session = 1;
	current->pgrp = 1;
	sigfillset(&current->blocked);
	current->fs->umask = 0;
	strcpy(current->comm,"IR poll");

	ir->thread = current;

#ifdef __SMP__
	unlock_kernel();
#endif

	printk("ir: thread started\n");
	if(ir->notify != NULL)
		up(ir->notify);
		
	for (;;) {
		if (ir->rmmod)
			break;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/10);
		if (signal_pending(current))
			break;

		/* poll IR chip */
		if (3 != i2c_master_recv(client,b,3)) {
			dprintk("ir: read error\n");
			continue;
		}

		/* key pressed ? */
		if (b[0] != REPEAT_TOGGLE_0 && b[0] != REPEAT_TOGGLE_1)
			continue;
		
		/* look what we have */
		dprintk("ir: key %s (0x%02x/0x%02x)\n", keyname(b[1]), b[0], b[1]);
		if (b[0] == last_b0) { /* repeat */
			dprintk("Repeat\n");
		} else {
			last_b0 = b[0];
		}
		repeat_bit=(b[0]&0x20) ? 0x800:0;
		code = (0x1000 | repeat_bit | (b[1]>>2));
		
		/* put into fifo */
		down(&ir->lock);
		if (((ir->tail+1)%BUFLEN) == ir->head) {
			up(&ir->lock);
			dprintk("ir: input buffer overflow\n");
		} else {
			ir->buffer[ir->tail++]=code;
			ir->tail%=BUFLEN;
			up(&ir->lock);
			wake_up_interruptible(&ir->wait_poll);
		}
	}

	dprintk("ir: thread done\n");
	ir->thread = NULL;

	if(ir->notify != NULL)
		up(ir->notify);
	return 0;
}

/* ----------------------------------------------------------------------- */

static unsigned int lirc_haup_poll(struct file *file,
				   struct poll_table_struct * wait)
{
	struct i2c_client *client = file->private_data;
	struct IR *ir = client->data;

        poll_wait(file, &ir->wait_poll, wait);

	if(ir->head != ir->tail)
		return POLLIN | POLLRDNORM;

	return 0;
}

static int lirc_haup_ioctl(struct inode *ino, struct file *fp,
			   unsigned int cmd, unsigned long arg)
{
        int result = 0;
	unsigned long features    = LIRC_CAN_REC_LIRCCODE;
	unsigned long mode;
	
	switch (cmd) {
	case LIRC_GET_FEATURES:
		result = put_user(features,(unsigned long *) arg);
		break;
	case LIRC_SET_REC_MODE:
		result = get_user(mode, (unsigned long *) arg);
		if (result) {
			break;
		}
		if ((LIRC_MODE2REC(mode) & features) == 0) {
			result = -EINVAL;
		}
		break;
	case LIRC_GET_LENGTH:
		result = put_user(CODE_LENGTH, (unsigned long *)arg);
		break;
	default:
		result = -ENOIOCTLCMD;
	}
	return result;
}

/* This function is called whenever a process attempts to open the device
 * file */
static int lirc_haup_open(struct inode *inode, struct file *file)
{
	struct i2c_client *client = irs[0];
	struct IR *ir;

	if (NULL == client)
		return -ENODEV;
	file->private_data = client;
	ir = client->data;
	
	down(&ir->lock);
	if (0 == ir->open) {
		ir->head = 0;
		ir->tail = 0;
	}
	ir->open++;
	up(&ir->lock);

	/* lock bttv in memory while /dev/lirc is in use  */
	if (client->adapter->inc_use) 
		client->adapter->inc_use(client->adapter);

	MOD_INC_USE_COUNT;
	return 0;
}

/* This function is called when a process closes the device file. */
static int lirc_haup_close(struct inode *inode, struct file *file)
{
	struct i2c_client *client = file->private_data;
	struct IR *ir = client->data;

	ir->open--;

	if (client->adapter->dec_use) 
		client->adapter->dec_use(client->adapter);
	MOD_DEC_USE_COUNT;
	return 0;
}

static ssize_t lirc_haup_read(struct file * file, char * buffer,
			      size_t length, loff_t *ppos)
{
	struct i2c_client *client = file->private_data;
	struct IR *ir = client->data;
        unsigned char data[2];

	down(&ir->lock);

	/* we are only prepared to send 2 bytes */
	if(length != 2) {
		up(&ir->lock);
		return -EIO;
	}

	/* while input buffer is empty, wait for input */
	while (ir->head==ir->tail) {
		up(&ir->lock);
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (signal_pending(current))
			return -ERESTARTSYS;
		interruptible_sleep_on(&ir->wait_poll);
		current->state = TASK_RUNNING;
		down(&ir->lock);
	}
	
	data[0]=(unsigned char) ((ir->buffer[ir->head]>>8)&0x1f);
	data[1]=(unsigned char) (ir->buffer[ir->head++]&0xff);
	ir->head%=BUFLEN;
	up(&ir->lock);
	
	if (copy_to_user(buffer,data,2))
	        return -EFAULT;
	return length;
}


static ssize_t lirc_haup_write(struct file * file, const char * buffer,
			       size_t length, loff_t *ppos)
{
	return -EINVAL;
}

static struct file_operations lirc_haup_fops = {
	read:		lirc_haup_read,
	write:		lirc_haup_write,
	poll:		lirc_haup_poll,
	ioctl:		lirc_haup_ioctl,
	open:		lirc_haup_open,
	release:	lirc_haup_close
};

/* ----------------------------------------------------------------------- */

static int ir_attach(struct i2c_adapter *adap, int addr,
		      unsigned short flags, int kind);
static int ir_detach(struct i2c_client *client);
static int ir_probe(struct i2c_adapter *adap);
static int ir_command(struct i2c_client *client, unsigned int cmd, void *arg);

static struct i2c_driver driver = {
        "i2c ir driver",
        /* I2C_DRIVERID_FIXME */ 42,
        I2C_DF_NOTIFY,
        ir_probe,
        ir_detach,
        ir_command,
};

static struct i2c_client client_template = 
{
        "ir",
        -1,
        0,
        0,
        NULL,
        &driver
};

static int ir_attach(struct i2c_adapter *adap, int addr,
		     unsigned short flags, int kind)
{
	int i;
	DECLARE_MUTEX_LOCKED(sem);
	struct IR *ir;
        struct i2c_client *c;

        client_template.adapter = adap;
        client_template.addr = addr;

        if (NULL == (c = kmalloc(sizeof(struct i2c_client),GFP_KERNEL)))
                return -ENOMEM;
        memcpy(c,&client_template,sizeof(struct i2c_client));
	if (NULL == (ir = kmalloc(sizeof(struct IR),GFP_KERNEL))) {
		kfree(c);
		return -ENOMEM;
	}
	
	memset(ir,0,sizeof(struct IR));
        init_MUTEX(&ir->lock);
	init_waitqueue_head(&ir->wait_poll);
	c->data = ir;

	/* startup thread */
	ir->notify = &sem;
	kernel_thread(ir_poll_thread, (void *)c, 0);
	down(&sem);
	ir->notify = NULL;

	/* register device */
	register_chrdev(LIRC_MAJOR, DEVICE_NAME, &lirc_haup_fops);

	/* update our own array */
	for (i = 0; i < IR_MAX; i++) {
		if (NULL == irs[i]) {
			irs[i] = c;
			break;
		}
	}
	
	/* done */
        i2c_attach_client(c);
	return 0;
}

static int ir_detach(struct i2c_client *client)
{
	DECLARE_MUTEX_LOCKED(sem);
	struct IR *ir  = (struct IR*)client->data;
	int i;
	
	/* unregister device */
	unregister_chrdev(LIRC_MAJOR, DEVICE_NAME);

	/* shutdown control thread */
	if (ir->thread) 
	{
		ir->notify = &sem;
		ir->rmmod = 1;
		down(&sem);
		ir->notify = NULL;
	}

        /* update our own array */
	for (i = 0; i < IR_MAX; i++) {
		if (client == irs[i]) {
			irs[i] = NULL;
			break;
		}
	}

	i2c_detach_client(client);
	kfree(ir);
	kfree(client);
	return 0;
}

static int ir_probe(struct i2c_adapter *adap)
{
	if (adap->id == (I2C_ALGO_BIT | I2C_HW_B_BT848))
		return i2c_probe(adap, &addr_data, ir_attach);
	return 0;
}

static int ir_command(struct i2c_client *client,unsigned int cmd, void *arg)
{
	/* nothing */
	return 0;
}

/* ----------------------------------------------------------------------- */

#ifdef MODULE
int init_module(void)
#else
int ir_init(void)
#endif
{
	i2c_add_driver(&driver);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	i2c_del_driver(&driver);
}
#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
