/*
 * The Kernel module driver for the PIC16C54 chip, found on ProVision PV951 TV/FM cards
 * This driver provides the audio and lirc services.
 *
 * Copyright (c) 2000 Kefah T. Issa <Kefah.Issa@usa.net>
 * This code is placed under the terms of the GNU General Public License.
 *
 * In writting this code, I relied on tea6300 and ir drivers.
 *
 * Thanks to:
 * ProVision (Arrow Hsu <arrow@provision.com.tw>)
 *   for the card access information, specifications, and support.
 * Mark D. Studebaker <mds@paradyne.com> 
 *   for the valiable I2C bus speed control tip.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/videodev.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <asm/semaphore.h>

#ifdef __SMP__
#include <asm/pgtable.h>
#include <linux/smp_lock.h>
#endif
/* kernel_thread */
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>


#include "bttv.h"
#include "audiochip.h"
#include "pic16c54.h"
#include "lirc.h"

#define DEVICE_NAME 	"ProVedio 951 ir rc"
#define LIRC_MAJOR  	61
#define CODE_LENGTH		8


// Define UDELAY to control i2c bus speed.
// 84 ===> 84 microseonds for half a signal ==> frequency = 1 / ( 84e-6 * 2) ~= 6KHz
#define	UDELAY	84

#define SUCCESS 0

#define BUFLEN 256
#define IR_MAX 4

struct IR
{
	struct task_struct  *thread;
	struct semaphore    *notify;
	int                 rmmod;
	int                 open;

	struct semaphore    lock;
        wait_queue_head_t   wait_poll;
	unsigned char buffer[BUFLEN];
	int head,tail;
};

static struct i2c_client *irs[IR_MAX];


/* Addresses to scan */
static unsigned short normal_i2c[] = {ID_16C54 >> 1,  I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {I2C_CLIENT_END};
static unsigned short probe[2]        = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short probe_range[2]  = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore[2]       = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short ignore_range[2] = { I2C_CLIENT_END, I2C_CLIENT_END };
static unsigned short force[2]        = { I2C_CLIENT_END, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data =
{
	normal_i2c, normal_i2c_range, 
	probe, probe_range, 
	ignore, ignore_range, 
	force
};


static int debug = 0; /* insmod parameter */
MODULE_PARM(debug,"i");

#define dprintk  if (debug) printk

static struct i2c_driver driver;
static struct i2c_client client_template;


/* ********************************* *
 * functions for talking to PIC16C54 *
 * ********************************* */

static int pic16c54_write(struct i2c_client *client, int addr, int val)
{
	unsigned char buffer[2];
	int ret=0;
	struct i2c_algo_bit_data* algo_data = client->adapter->algo_data;
	int old_udelay = algo_data->udelay;
	algo_data->udelay = UDELAY;
	
	buffer[0] = addr;
	buffer[1] = val;
	
	if(2 != i2c_master_send(client,buffer,2))
	{
		printk(KERN_WARNING "PIC16C54: I/O error, trying (write %d 0x%x)\n",addr, val);
		ret=-1;
	}else
	{
		ret = 0;
	}

	algo_data->udelay = old_udelay;

	return ret;
}

static int pic16c54_read(struct i2c_client *client, unsigned char* buffer, int length)
{
	int ret=0;
	struct i2c_algo_bit_data* algo_data = client->adapter->algo_data;
	int old_udelay = algo_data->udelay;
	algo_data->udelay = UDELAY;
	
	if(length != i2c_master_recv(client,buffer,length))
	{
		printk(KERN_WARNING "PIC16C54: I/O error, trying (read %d bytes)\n",length);
		ret=-1;
	}else
	{
		ret = length;
	}

	algo_data->udelay = old_udelay;

	return ret;
}



/* ****************** *
 * IR file operations *
 * ****************** */
static int ir_poll_thread(void *data)
{
	struct i2c_client *client = data;
	struct IR *ir = client->data;
    unsigned char b;


#ifdef __SMP__
	lock_kernel();
#endif

	exit_mm(current);
	current->session = 1;
	current->pgrp = 1;
	sigfillset(&current->blocked);
	current->fs->umask = 0;
	strcpy(current->comm,"PIC16C54 - IR poll");

	ir->thread = current;

#ifdef __SMP__
	unlock_kernel();
#endif

	printk("ir: thread started\n");
	if(ir->notify != NULL)
		up(ir->notify);
		
	for (;;)
	{
		if (ir->rmmod)
			break;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ/10);
		if (signal_pending(current))
			break;

		/* poll IR chip */
		if (1 != pic16c54_read(client,&b,1))
		{
			dprintk("ir: read error\n");
			continue;
		}

		/* ignore 0xaa */
		if(b == 0xaa)
			continue;
		
		/* look what we have */
		dprintk("ir: key 0x%02x)\n", b);
		
		/* put into fifo */
		down(&ir->lock);
		if (((ir->tail+1)%BUFLEN) == ir->head)
		{
			up(&ir->lock);
			dprintk("ir: input buffer overflow\n");
		} else
		{
			ir->buffer[ir->tail++]=b;
			ir->tail%=BUFLEN;
			up(&ir->lock);
			wake_up_interruptible(&ir->wait_poll);
		}
	}

	dprintk("ir: thread done\n");
	ir->thread = NULL;

	if(ir->notify != NULL)
		up(ir->notify);
	return 0;
}

/* ----------------------------------------------------------------------- */

static unsigned int lirc_poll(struct file *file, struct poll_table_struct * wait)
{
	struct i2c_client *client = file->private_data;
	struct IR *ir = client->data;

        poll_wait(file, &ir->wait_poll, wait);

	if(ir->head != ir->tail)
		return POLLIN | POLLRDNORM;

	return 0;
}


static int lirc_ioctl(struct inode *ino, struct file *fp,unsigned int cmd, unsigned long arg)
{
    int result = 0;
	unsigned long features    = LIRC_CAN_REC_LIRCCODE;
	unsigned long mode;
	
	switch (cmd)
	{
	case LIRC_GET_FEATURES:
		result = put_user(features,(unsigned long *) arg);
		break;
	case LIRC_SET_REC_MODE:
		result = get_user(mode, (unsigned long *) arg);
		if (result)
		{
			break;
		}
		if ((LIRC_MODE2REC(mode) & features) == 0)
		{
			result = -EINVAL;
		}
		break;
	case LIRC_GET_LENGTH:
		result = put_user(CODE_LENGTH, (unsigned long *)arg);
		break;
	default:
		result = -ENOIOCTLCMD;
	}
	return result;
}

/* This function is called whenever a process attempts to open the device
 * file */
static int lirc_open(struct inode *inode, struct file *file)
{
	struct i2c_client *client = irs[0];
	struct IR *ir;

	if (NULL == client)
		return -ENODEV;
	file->private_data = client;
	ir = client->data;
	
	down(&ir->lock);
	if (0 == ir->open)
	{
		ir->head = 0;
		ir->tail = 0;
	}
	ir->open++;
	up(&ir->lock);

	/* lock bttv in memory while /dev/lirc is in use  */
	if (client->adapter->inc_use)
		client->adapter->inc_use(client->adapter);

	MOD_INC_USE_COUNT;
	return SUCCESS;
}


/* This function is called when a process closes the device file. */
static int lirc_close(struct inode *inode, struct file *file)
{
	struct i2c_client *client = file->private_data;
	struct IR *ir = client->data;

	ir->open--;

	if (client->adapter->dec_use)
		client->adapter->dec_use(client->adapter);
		
	MOD_DEC_USE_COUNT;
	return SUCCESS;
}

static ssize_t lirc_read(struct file * file, char * buffer,size_t length, loff_t *ppos)
{
	struct i2c_client *client = file->private_data;
	struct IR *ir = client->data;
    unsigned char data;

	down(&ir->lock);

	/* we are only prepared to send 1 byte */
	if(length != 1)
	{
		up(&ir->lock);
		return -EIO;
	}

	/* while input buffer is empty, wait for input */
	while (ir->head==ir->tail)
	{
		up(&ir->lock);
		if (file->f_flags & O_NONBLOCK)
			return -EAGAIN;
		if (signal_pending(current))
			return -ERESTARTSYS;
		interruptible_sleep_on(&ir->wait_poll);
		current->state = TASK_RUNNING;
		down(&ir->lock);
	}
	
	data = ir->buffer[ir->head++];
	
	ir->head%=BUFLEN;
	up(&ir->lock);
	
	if (copy_to_user(buffer,&data,1))
	        return -EFAULT;
	return length;
	
}



static ssize_t lirc_write(struct file * file, const char * buffer,size_t length, loff_t *ppos)
{
	return -EINVAL;
}


static struct file_operations lirc_pv951_fops =
{
	NULL,		// seek
	lirc_read,	// read
	lirc_write, // write
	NULL,       // readdir
	lirc_poll,	// poll
	lirc_ioctl, // ioctl
	NULL,       // mmap
	lirc_open,  // open
	NULL,       // flush
	lirc_close  // release
};



/* *********************** *
 * i2c interface functions *
 * *********************** */

static int pic16c54_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	
	if(cmd == AUDC_SET_INPUT)    /* Interpret messages from bttv */
	{
		__u16* sarg = arg;

		switch(*sarg)
		{
		case AUDIO_EXTERN:
			pic16c54_write(client, REG_MISC, MISC_SND_NOTMUTE);		
			pic16c54_write(client, REG_MISC, MISC_SWITCH_LINE_IN);
			break;
		
		case AUDIO_TUNER:
		case AUDIO_RADIO:
			pic16c54_write(client, REG_MISC, MISC_SND_NOTMUTE);		
			pic16c54_write(client, REG_MISC, MISC_SWITCH_TUNER);
			break;
		
		case AUDIO_OFF:
			pic16c54_write(client, REG_MISC, MISC_SND_MUTE);		
			break;
		
		case AUDIO_ON:
		default:
			pic16c54_write(client, REG_MISC, MISC_SND_NOTMUTE);		
			break;
		}		
	}
	
	return SUCCESS;
}


static int pic16c54_attach(struct i2c_adapter *adap, int addr, unsigned short flags, int kind)
{
	
	int i;
	DECLARE_MUTEX_LOCKED(sem);
	struct IR *ir;
	struct i2c_client *client;

    client_template.adapter = adap;
    client_template.addr = addr;
	
	dprintk(KERN_DEBUG "pic16c54_attach\n");


    if (NULL == (client = kmalloc(sizeof(struct i2c_client),GFP_KERNEL)))
          return -ENOMEM;

    memcpy(client,&client_template,sizeof(struct i2c_client));
	
	if (NULL == (ir = kmalloc(sizeof(struct IR),GFP_KERNEL))) {
		kfree(client);
		return -ENOMEM;
	}
	
	memset(ir,0,sizeof(struct IR));
    init_MUTEX(&ir->lock);
	init_waitqueue_head(&ir->wait_poll);
	client->data = ir;

	/* startup thread */
	ir->notify = &sem;
	kernel_thread(ir_poll_thread, (void *)client, 0);
	down(&sem);
	ir->notify = NULL;

	/* register device */
    register_chrdev(LIRC_MAJOR, DEVICE_NAME, &lirc_pv951_fops);

	/* update our own array */
	for (i = 0; i < IR_MAX; i++)
	{
		if (NULL == irs[i])
		{
			irs[i] = client;
			break;
		}
	}
	
	pic16c54_write(client, REG_MISC, MISC_SWITCH_TUNER); /* Switch audio to tuner*/
	pic16c54_write(client, REG_MISC, MISC_SND_MUTE);
	

	MOD_INC_USE_COUNT;
	
	strcpy(client->name,"PIC16C54");
	printk(KERN_INFO "pic16c54: initialized\n");

	i2c_attach_client(client);
	return SUCCESS;
}

static int pic16c54_probe(struct i2c_adapter *adap)
{
	int ret = 1;
	struct i2c_algo_bit_data* algo_data = adap->algo_data;
	int old_udelay = algo_data->udelay;
	algo_data->udelay = UDELAY;
	
	dprintk(KERN_DEBUG "pic16c54_probe \n");
	
	if (adap->id == (I2C_ALGO_BIT | I2C_HW_B_BT848))
	{
		ret =  i2c_probe(adap, &addr_data, pic16c54_attach);
		if(ret!=0)
		{
			dprintk(KERN_WARNING "PIC16C54: I/O Error : probing faild!");
		}
	}

	algo_data->udelay = old_udelay;
	
	return ret;
}

static int pic16c54_detach(struct i2c_client *client)
{
	DECLARE_MUTEX_LOCKED(sem);
	struct IR *ir  = (struct IR*)client->data;
	int i;

	dprintk(KERN_DEBUG "pic16c54_detach\n");
		
	/* unregister device */
	unregister_chrdev(LIRC_MAJOR, DEVICE_NAME);

	/* shutdown control thread */
	if (ir->thread)
	{
		ir->notify = &sem;
		ir->rmmod = 1;
		down(&sem);
		ir->notify = NULL;
	}

     /* update our own array */
	for (i = 0; i < IR_MAX; i++)
	{
		if (client == irs[i])
		{
			irs[i] = NULL;
			break;
		}
	}
	
	pic16c54_write(client, REG_MISC, MISC_SWITCH_TUNER); /* Switch audio to tuner*/
	pic16c54_write(client, REG_MISC, MISC_SND_MUTE);

	i2c_detach_client(client);
	
	kfree(ir);
	kfree(client);
	
	MOD_DEC_USE_COUNT;
	return SUCCESS;
}



static struct i2c_driver driver =
{
        "i2c pic16c54 driver",
        I2C_DRIVERID_TEA6300, /* Should be changed!*/
        I2C_DF_NOTIFY,
		pic16c54_probe,
        pic16c54_detach,
        pic16c54_command,
};

static struct i2c_client client_template =
{
        "(unset)",		/* name */
        -1,
        0,
        0,
        NULL,
        &driver
};

int init_module(void)
{
	i2c_add_driver(&driver);
	return SUCCESS;
}

void cleanup_module(void)
{
	i2c_del_driver(&driver);
}


