/*
 * linux/fs/nls_euc-jp.c
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/nls.h>
#include <linux/init.h>
#include <linux/errno.h>

static struct nls_table *p_nls = NULL;


#define SS2		(0x8E)		/* Single Shift 2 */
#define SS3		(0x8F)		/* Single Shift 3 */

static int uni2char(unsigned char ch, unsigned char cl, unsigned char *out, int boundlen, int *outlen)
{
	if (p_nls && p_nls->uni2char(ch, cl, out, boundlen, outlen) < 0)
		return -EINVAL;
	
	/* translate SJIS into EUC-JP */
	if (*outlen == 1) {
		/* JIS X 201 KANA */
		if (0xA1 <= out[0] && out[0] <= 0xDF) {
			if (boundlen <= 1)
				return -EINVAL;
			out[1] = out[0];
			out[0] = SS2;
			*outlen = 2;    
		}
	} else if (*outlen == 2) { 
		/* JIS X 208 */

		/* SJIS codes 0xF0xx to 0xFFxx are machine-dependent codes and user-defining characters */
		if (out[0] >= 0xF0) {
			out[0] = 0x81;  /* 'GETA' with SJIS coding */
			out[1] = 0xAC;
		}

		out[0] = (out[0]^0xA0)*2 + 0x5F;
		if (out[1] > 0x9E)
			out[0]++; 
		
		if (out[1] < 0x7F)
			out[1] = out[1] + 0x61;
		else if (out[1] < 0x9F)
			out[1] = out[1] + 0x60;
		else
			out[1] = out[1] + 0x02;
	}
	else
		return -EINVAL;

	return *outlen;
}

static int char2uni(unsigned char *rawstring, int boundlen, int *offset,
				unsigned char *uni1, unsigned char *uni2)
{
	unsigned char sjis_temp[2];
	int euc_offset, n;
	
	if (!p_nls)
		return -EINVAL;
	if (boundlen <= 0)
		return -ENAMETOOLONG;

	if (boundlen == 1){
		*uni1 = rawstring[0];
		*uni2 = 0x00;
		return 1;
	}

	/* translate EUC-JP into SJIS */
	if (rawstring[0] > 0x7F) {
		if (rawstring[0] == SS2) {
			/* JIS X 201 KANA */
			sjis_temp[0] = rawstring[1];
			sjis_temp[1] = 0x00;
			euc_offset = 2;
		} else if (rawstring[0] == SS3) {
			/* JIS X 212 */
			sjis_temp[0] = 0x81; /* 'GETA' with SJIS coding */
			sjis_temp[1] = 0xAC;
			euc_offset = 3;
		} else { 
			/* JIS X 208 */
			sjis_temp[0] = ((rawstring[0]-0x5f)/2) ^ 0xA0;
			if (!(rawstring[0]&1))
				sjis_temp[1] = rawstring[1] - 0x02;
			else if (rawstring[1] < 0xE0)
				sjis_temp[1] = rawstring[1] - 0x61;
			else
				sjis_temp[1] = rawstring[1] - 0x60;
			euc_offset = 2;
		}
	} else { 
		/* JIS X 201 ROMAJI */
		sjis_temp[0] = rawstring[0];
		sjis_temp[1] = rawstring[1];
		euc_offset = 1;
	}

	if ( (n = p_nls->char2uni(sjis_temp, boundlen, offset, uni1, uni2)) < 0)
		return n;

	*offset = euc_offset;
	return *offset;
}

static struct nls_table table = {
	"euc-jp",
	uni2char,
	char2uni,
	NULL,
	NULL,
	NULL
};

int __init init_nls_euc_jp(void)
{
	p_nls = load_nls("cp932");

	if (p_nls){
		table.inc_use_count = p_nls->inc_use_count;
		table.dec_use_count = p_nls->dec_use_count;
		return register_nls(&table);
	}

	return -EINVAL;
}

#ifdef MODULE
int init_module(void)
{
	return init_nls_euc_jp();
}


void cleanup_module(void)
{
	unregister_nls(&table);
	unload_nls(p_nls);
}
#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 *
---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -8
 * c-argdecl-indent: 8
 * c-label-offset: -8
 * c-continued-statement-offset: 8
 * c-continued-brace-offset: 0
 * End:
 */
